"""
Some plugins to support the tests.
"""

from __future__ import print_function

import random

import caparg as ca
import elcaminoreal


COMMANDS = elcaminoreal.Commands()


@COMMANDS.dependency(name="foo", dependencies=["bar"])
def a_foo(dependencies, _possible_dependencies):
    """
    Depend on a bar object
    """
    return dict(bar=dependencies['bar'])


@COMMANDS.dependency(dependencies=["bar", "quux"],
                     possible_dependencies=["foo"],
                     regular=True)
def regular(bar, quux, build_foo):
    """
    Depend on bar, maybe on foo

    Use regular arguments.
    """
    return dict(bar=bar, quux=quux, foo=build_foo())


@COMMANDS.dependency(possible_dependencies=["bar"])
def foo_2(_dependencies, possible_dependencies):
    """
    Depend on a bar object in an optional way.
    """
    return dict(bar=possible_dependencies['bar']())


@COMMANDS.dependency(name="bar")
def a_bar(_dependencies, _possible_dependencies):
    """
    Return a bar-like object.
    """
    return "I'm a bar"


@COMMANDS.dependency(name="quux")
def a_quux(_dependencies, _possible_dependencies):
    """
    Return a quux-like object.
    """
    return "I'm a quux"


@COMMANDS.dependency()
def rand(_dependencies, _possible_dependencies):
    """
    Generate a random number.
    """
    return random.random()


@COMMANDS.dependency(dependencies=["rand"])
def needs_rand(dependencies, _possible_dependencies):
    """
    Depend on a random number.
    """
    return dict(rand=dependencies["rand"])


@COMMANDS.dependency(name="baz")
def a_baz(dependencies, _possible_dependencies):
    """
    Use an undeclared dependency.
    """
    return dependencies['bar']


@COMMANDS.dependency(dependencies=['tuck'])
def robin(_dependencies, _possible_dependencies):
    """
    Depend on tuck
    """


@COMMANDS.dependency(dependencies=['robin'])
def tuck(_dependencies, _possible_dependencies):
    """
    Depend on robin
    """


@COMMANDS.dependency(name='print')
def _print(_dependencies, _possible_dependencies):
    """
    Return a function to display things on the terminal.
    """
    return print


@COMMANDS.dependency(name='output')
def dummy_output(_dependencies, _possible_dependencies):
    """
    Literally do nothing.

    This is designed for being overridden.
    """


@COMMANDS.command(dependencies=['foo', 'output'],
                  parser=ca.command('',
                                    ca.positional('lili', type=str)),
                  regular=True)
def regular_command(foo, lili, output):
    """
    Use regular arguments

    Output results
    """
    output(foo, lili)


@COMMANDS.command(dependencies=['foo', 'print'],
                  parser=ca.command('',
                                    ca.positional('lala', type=str)))
def show(args, dependencies):
    """
    Print then arguments.
    """
    dependencies['print'](args, "BREAK", dependencies)


@COMMANDS.command(dependencies=['bar', 'print'],
                  parser=ca.command('',
                                    wooo=ca.option(type=str)))
def gowoo(args, dependencies):
    """
    Print 'woo' and then arguments.
    """
    dependencies['print']("woo", args, dependencies)


@COMMANDS.command(dependencies=['print'],
                  parser=ca.command('',
                                    foo=ca.option(type=str),
                                    bar=ca.option(type=str, required=True)))
def interesting_args(args, dependencies):
    """
    Print arguments.
    """
    dependencies['print'](args.get('foo'), args['bar'])



"""
El Camino Real Test Main

Run the built-in test El Camino Real plugins.
"""
import sys
from elcaminoreal.test import some_plugins

if __name__ != '__main__':
    raise ImportError("main module, not importable", __name__)

some_plugins.COMMANDS.run(sys.argv[1:])



"""
Test dependencies and commands
"""
from __future__ import print_function

import unittest

import six

import elcaminoreal
from elcaminoreal.test import some_plugins


class DependencyResolverTester(unittest.TestCase):

    """
    Dependency resolution tests
    """

    def test_mkgraph(self):
        """
        mkgraph finds and runs dependencies without subdependencis
        """
        result = some_plugins.COMMANDS.mkgraph(['bar'])
        self.assertEquals(result.pop('bar'), "I'm a bar")
        self.assertEquals(result, {})

    def test_mkgraph_cycle(self):
        """
        mkgraph raises ValueError when it detects a cycle
        """
        with self.assertRaises(ValueError):
            some_plugins.COMMANDS.mkgraph(['robin'])

    def test_mkgraph_random(self):
        """
        mkgraph caches subdependencies
        """
        result = some_plugins.COMMANDS.mkgraph(['rand', 'needs_rand'])
        self.assertEquals(result['rand'], result['needs_rand']['rand'])

    def test_mkgraph_possible(self):
        """
        mkgraph possible dependencies are functions that return the object
        """
        result = some_plugins.COMMANDS.mkgraph(['foo_2'])
        self.assertEquals(result['foo_2'], dict(bar="I'm a bar"))

    def test_mkgraph_regular(self):
        """
        mkgraph regular functions work
        """
        result = some_plugins.COMMANDS.mkgraph(['regular'])
        self.assertEquals(result['regular']['bar'], result['bar'])
        self.assertEquals(result['regular']['quux'], result['quux'])
        self.assertEquals(result['regular']['foo'], result['foo'])


class RunnerResolverTester(unittest.TestCase):

    """
    Command running tests
    """

    def test_run(self):
        """
        Run calls the command with the dependencies it requested
        """
        output = []

        def _my_print(*args):
            output.append(' '.join(map(str, args)))
        some_plugins.COMMANDS.run(['show', 'heee'],
                                  override_dependencies=dict(print=_my_print))
        self.assertEquals(len(output), 1)
        args, deps = output[0].split('BREAK', 1)
        self.assertIn('hee', args)
        self.assertIn('foo', deps)
        self.assertIn('bar', deps)

    def test_args(self):
        """
        Argument parsing respects required/not-required
        """
        output = []

        def _my_print(*args):
            output.append(' '.join(map(str, args)))
        some_plugins.COMMANDS.run(['interesting-args', '--bar', 'ddd'],
                                  override_dependencies=dict(print=_my_print))
        self.assertEquals(len(output), 1)
        my_foo, my_bar = output.pop().split(' ')
        self.assertEquals(my_foo, 'None')
        self.assertEquals(my_bar, 'ddd')

    def test_required(self):
        """
        Argument parsing without required arguments fails`
        """
        with self.assertRaises(BaseException):
            some_plugins.COMMANDS.run(['interesting_args', '--foo', 'ddd'])

    def test_error_redirect(self):
        """
        Error redirect sends error message to file pointer
        """
        filep = six.moves.StringIO('w')
        with elcaminoreal.errors_to(filep):
            some_plugins.COMMANDS.run(['no-such-command'])
        error_message = filep.getvalue().splitlines()
        self.assertEquals(error_message.pop(0), 'Usage:')

    def test_regular(self):
        """
        Asking for regular arguments calls functions with argument names
        """
        output = []

        def _my_output(*args):
            output.append(args)
        dependencies = dict(output=_my_output)
        some_plugins.COMMANDS.run(['regular-command', 'thing'],
                                  override_dependencies=dependencies)
        self.assertEquals(len(output), 1)
        self.assertEquals(output[0][0]['bar'], "I'm a bar")
        self.assertEquals(output[0][1], 'thing')



"""
Gather dependencies and commands
"""
import functools

import attr
import caparg
import gather
import pyrsistent

_COLLECTOR_FACTORY = attr.Factory(functools.partial(gather.Collector, depth=1))


@attr.s(frozen=True)
class Commands(object):

    """
    A command and dependency gatherer.
    """

    _collector = attr.ib(default=_COLLECTOR_FACTORY, init=False)
    _command_collector = attr.ib(default=_COLLECTOR_FACTORY, init=False)

    def command(self,
                name=None,
                parser=caparg.command(''),
                dependencies=pyrsistent.v(),
                regular=False):
        """
        Register as a command.

        """
        transform = gather.Wrapper.glue((dependencies, parser, regular))
        ret = self._command_collector.register(name, transform=transform)
        return ret

    def run(self, args, override_dependencies=pyrsistent.m()):
        """
        Run a command

        """
        collection = {name.replace('_', '-'): value
                      for name, value in
                      self._command_collector.collect().items()}
        subcommands = [thing.extra[1].rename(name)
                       for name, thing in collection.items()]
        command = caparg.command('', *subcommands)
        parsed = command.parse(args)
        subcommand = ' '.join(parsed['__caparg_subcommand__'])
        func = collection[subcommand].original
        dependencies, _ignored, regular = collection[subcommand].extra
        graph = self.mkgraph(dependencies)
        graph.update(override_dependencies)
        if not regular:
            return func(parsed, graph)
        args = {dependency: graph[dependency]
                for dependency in dependencies}
        args.update(parsed)
        del args['__caparg_subcommand__']
        return func(**args)

    def dependency(self,
                   name=None,
                   dependencies=pyrsistent.v(),
                   possible_dependencies=pyrsistent.v(),
                   regular=False):
        """
        Register as a dependency.

        """
        glue = (dependencies, possible_dependencies, regular)
        transform = gather.Wrapper.glue(glue)
        ret = self._collector.register(name, transform=transform)
        return ret

    # Recursive implementation for now
    def mkgraph(self, things):
        """
        Resolve dependencies and generate them

        """
        collection = self._collector.collect()
        ret = {}

        def _build(thing, on_route=pyrsistent.s()):
            if thing in on_route:
                raise ValueError("circular dependency detected",
                                 thing, on_route)
            if thing in ret:
                return ret[thing]
            on_route = on_route.add(thing)
            plugin = collection[thing]
            func = plugin.original
            dependencies, possible_dependencies, regular = plugin.extra
            my_dependencies, my_possible_dependencies = {}, {}
            for other_thing in dependencies:
                my_dependencies[other_thing] = _build(other_thing, on_route)
            for other_thing in possible_dependencies:
                builder = functools.partial(_build, other_thing, on_route)
                my_possible_dependencies[other_thing] = builder
            if regular:
                args = {'build_' + key: value
                        for key, value in my_possible_dependencies.items()}
                args.update(my_dependencies)
                ret[thing] = func(**args)
            else:
                ret[thing] = func(my_dependencies, my_possible_dependencies)
            return ret[thing]
        for thing in things:
            _build(thing)
        return ret



"""
Skeleton for Voynich

Voynich is a system for encrypting secrets.
This is a rough skeleton, to show-case how elcaminoreal
would work for a real command.
"""
import os
import sys

import caparg as ca
import elcaminoreal

COMMANDS = elcaminoreal.Commands()


@COMMANDS.dependency()
def current_directory(_deps, _maybedeps):
    """
    Current working directory
    """
    return '.'


@COMMANDS.dependency()
def environment(_deps, _maybedeps):
    """
    Environment dictionary
    """
    return os.environ


@COMMANDS.dependency(dependencies=['environment', 'current_directory'])
def secret_filename(dependencies, _maybedeps):
    """
    The filename to put the encrypted secrets in.
    """
    if 'VOYNICH_FILE' in dependencies['environment']:
        return dependencies['environment']['VOYNICH_FILE']
    return os.path.join(dependencies['current_directory'], 'voynich.json')


@COMMANDS.command(dependencies=['secret_filename'],
                  parser=ca.command('',
                                    key_file=ca.option(type=str,
                                                       required=True)))
def create(args, dependencies):
    """
    Create a new secrets file (and save the private key).
    """
    sys.stdout.write("writing key to {} and public data to {}\n".format(
        args.key_file,
        dependencies['secret_filename']))


@COMMANDS.command(dependencies=['secret_filename'],
                  parser=ca.command('',
                                    name=ca.option(type=str, required=True),
                                    value=ca.option(type=str, required=True)))
def encrypt(args, dependencies):
    """
    Add a new encrypted secret to the dependencies.
    """
    sys.stdout.write("adding to {} name {} value {}\n".format(
        dependencies['secret_filename'],
        args.name,
        args.value))


@COMMANDS.command(dependencies=['secret_filename'],
                  parser=ca.command('',
                                    key_file=ca.option(type=str,
                                                       required=True),
                                    directory=ca.option(type=str,
                                                        required=True)))
def decrypt(args, dependencies):
    """
    Decrypt the secrets from a file into a directory
    """
    sys.stdout.write("decrypting {} using {} into {}\n".format(
        dependencies['secret_filename'],
        args.key_file,
        args.directory))



"""
Run the pseduo-voynich examples.
"""
import sys

from elcaminoreal.test import voynich_skeleton

if __name__ != '__main__':
    raise ImportError("Module should not be imported", __name__)

voynich_skeleton.COMMANDS.run(sys.argv[1:])



"""
Capture argument parsing errors
"""
import contextlib

import caparg


@contextlib.contextmanager
def errors_to(filep):
    """
    A context manager for command-line parsing errors.

    It will catch caparg parsing errors, and redirect
    the output into the given file descriptor.

    Suggested usage:

    .. code::

        with errors_to(sys.stderr):
            COMMANDS.run(....)
    """
    try:
        yield
    except caparg.ParseError as exc:
        filep.write(exc.message)



import setuptools

with open('README.rst') as fp:
    long_description = fp.read()

setuptools.setup(
    name='elcaminoreal',
    license='MIT',
    description="El Camino Real: Bringing your things together.",
    long_description=long_description,
    use_incremental=True,
    setup_requires=['incremental'],
    author="Moshe Zadka",
    author_email="zadka.moshe@gmail.com",
    packages=setuptools.find_packages(where='src'),
    package_dir={"": "src"},
    install_requires=['attrs', 'incremental', 'gather', 'six', 'pyrsistent',
                      'caparg'],
    entry_points={
        'gather': [
             "gather=elcaminoreal",
        ]
    }
)
